#ifndef __LDRAWMODEL_H__
#define __LDRAWMODEL_H__

#include <stdio.h>

#include <LDLib/LDrawModelFile.h>
#include <LDLib/LDMError.h>
#include <LDLib/LDMFileLine.h>
#include <LDLib/Vector.h>
#include <LDLib/VectorShapeArrayTree.h>

#ifdef WIN32
#include <windows.h>
#endif // WIN32
#if defined(__APPLE__)
#	include <OpenGL/OpenGL.h>
#	include <GLUT/GLUT.h>
#	define APIENTRY
#else
#	include <GL/gl.h>
#endif

#include <TCFoundation/TCTypedObjectArray.h>
#include <TCFoundation/TCTypedDictionary.h>

class TGLShape;
class TGLEighthSphere;
class TGLTorus;
class TGLTorusQ;
class TCDictionary;
class TCStringArray;
class LDrawSubModel;
//class TCStringObjectTree;

extern const int LDM_QUAD_STRIPS;
extern const int LDM_QUADS;
extern const int LDM_TRIANGLES;
extern const int LDM_LINES;
extern const int LDM_LINE_STRIPS;
extern const int LDM_TRANS_LINES;
extern const int LDM_TRANS_LINE_STRIPS;

#define NUM_SPECIAL_COLORS 4

typedef int (*LDMProgressCallback)(char* message, float progress,
								   void* userData);
typedef int (*LDMErrorCallback)(LDMError* error, void* userData);

typedef void (APIENTRY *GLPolygonOffsetFunc)(GLfloat, GLfloat);

//typedef TCTypedObjectArray<TGLShape> TGLShapeArray;
typedef TCTypedObjectArray<TGLShapeArray> TCObjectArrayArray;
typedef TCTypedObjectArray<LDMFileLine> LDMFileLineArray;
typedef TCTypedObjectArray<LDrawSubModel> LDrawSubModelArray;
typedef TCTypedDictionary<TGLShapeArray> TGLShapeArrayDictionary;

typedef struct
{
	int colorNumber;
	float specular[4];
	float shininess;
} LDMSpecialColorSettings;

class LDrawModel: public LDrawModelFile
{
	public:
		LDrawModel(void);
		LDrawModel(LDrawModel&);
		LDrawModel(FILE*);

		LDrawModel* getParent(void) { return parent; }
		LDrawModel* getTopParent(void) { return topParent; }
		virtual int getNumCircleSegments(void);
		virtual int read(FILE* f, bool closeAfterRead, int subFileSize = -1);
		virtual bool draw(void);
		virtual bool draw(float* matrix, int colorNumber, int transparent);
		virtual bool draw(int shapeIndex, float* matrix, int colorNumber,
			int transparent, float progressFraction,
			float progressStartFraction);
		virtual bool draw(float*);
		virtual bool drawTransparent(float*);
//		virtual void setTransformation(int, float, float, float, float, float,
//			float, float, float, float, float, float, float);
		virtual void setTransformation(LDrawSubModel *subModel, float x,
			float y, float z, float a, float b, float c, float d, float e,
			float f, float g, float h, float i);
		virtual void setColor(int, char*);
		virtual void setColor(int, int, int, int, int);
//		int getColor(int index) { return colors[index]; }
//		void setColorNumber(int index, int num) { colorNumbers[index] = num; }
//		int getColorNumber(int index) { return colorNumbers[index]; }
		virtual void checkError(char*);
		virtual bool isTransparent(int);
		virtual void getMinMax(TGLShapeArray*, Vector&, Vector&, float*);
		virtual void getMinMax(Vector&, Vector&, float*);
		virtual void getMinMax(Vector&, Vector&);
		char* getName(void) { return name; }
		virtual void setName(char*);
		virtual void setFilename(char*);
		virtual void setCullBackFaces(int value);
		int getCullBackFaces(void) { return flags.cullBackFaces; }
		virtual bool compile(int = 0);
		virtual void uncompile(void);
//		virtual void computeChildren(void);
		void setAllowLowResStuds(bool value) { flags.allowLowResStuds = value; }
		void setAllowLowPrimitiveSubstitution(int value)
		{
			flags.allowPrimitiveSubstitution = value;
		}
		void setIsPart(int value) { flags.isPart = value; }
		int getCompiled(void) { return flags.compiled; }
		virtual void flatten(void);
		void setDrawHighlights(int value) { flags.drawHighlights = value; }
		int getDrawHighlights(void) { return flags.drawHighlights; }
		void setDrawConditionalHighlights(bool value)
		{
			flags.drawConditionalHighlights = value;
		}
		bool getDrawConditionalHighlights(void)
		{
			return flags.drawConditionalHighlights;
		}
		void setPerformSmoothing(bool value)
		{
			flags.performSmoothing = value;
		}
		bool getPerformSmoothing(void)
		{
			return flags.performSmoothing;
		}
		void setLineSmoothing(bool value)
		{
			flags.lineSmoothing = value;
		}
		bool getLineSmoothing(void)
		{
			return flags.lineSmoothing;
		}
		void setUseStipple(bool value) { flags.useStipple = value; }
		void setSortTransparent(int value) { flags.sortTransparent = value; }
		void setBlackHighlights(bool value) { flags.blackHighlights = value; }
		bool getBlackHighlights(void) { return flags.blackHighlights; }
//		int getUseStipple(void) { return flags.useStipple; }
		int getLineCount(void) { return getShapes(LDM_LINES)->getCount(); }
		int getLineStripCount(void)
		{
			return getShapes(LDM_LINE_STRIPS)->getCount();
		}
		int getQuadStripCount(void)
		{
			return getShapes(LDM_QUAD_STRIPS)->getCount();;
		}
		int getQuadCount(void) { return getShapes(LDM_QUADS)->getCount(); }
		int getTriangleCount(void)
		{
			return getShapes(LDM_TRIANGLES)->getCount();;
		}
		int getArrayCount(TCArray* array)
		{
			return array ? array->getCount(): 0;
		}
		virtual void writeToFile(void);
		virtual void writeToFile(FILE*, float*);
		virtual void writeShapeToFile(TGLShape*, FILE*, float*);
		virtual void writePointToFile(Vector &point, FILE *outputFile);
		virtual void setProgressCallback(LDMProgressCallback callback,
			void* userData);
		virtual void setErrorCallback(LDMErrorCallback callback,
			void* userData);
		void setCurveQuality(int value) { curveQuality = value; }
		int getCurveQuality(void) { return curveQuality; }
		void setCutawayDraw(bool value) { flags.cutawayDraw = value; }
		bool getCutawayDraw(void) { return flags.cutawayDraw; }
		virtual void setDefaultRGB(TCByte r, TCByte g, TCByte b);
		virtual void setDefaultColorNumber(int value);
		virtual TCObject *copy(void);
		bool getCompiling(void) { return flags.compiling; }

		static const char* lDrawDir(void);
		static void setUsePolygonOffset(int value) { usePolygonOffset = value; }
		static void setPolygonOffsetFunc(GLPolygonOffsetFunc func)
		{
			polygonOffsetFunc = func;
		}
		static GLPolygonOffsetFunc getPolygonOffsetFunc(void)
		{
			return polygonOffsetFunc;
		}
		static int getUsePolygonOffset(void) { return usePolygonOffset; }
		static void setSeamWidth(float value) { seamWidth = value; }
		static float getSeamWidth(void) { return seamWidth; }
		static void setDrawWireframe(int value) { drawWireframe = value; }
		static int getDrawWireframe(void) { return drawWireframe; }
		static void setUseLighting(int value) { useLighting = value; }
		static int getUseLighting(void) { return useLighting; }
		static void setLDrawDir(const char*);
//		static int getLastColor(void);
//		static void setLastColor(int); 
		static void getRGBA(int, int&, int&, int&, int&);

//		TCObject* retain(void);
	protected:
		virtual ~LDrawModel(void);
		virtual void dealloc(void);

//		virtual void compNormals(void);
//		virtual void compShapeNormals(TCObjectArray*);
		virtual void setParent(LDrawModel*);
		virtual void setTopParent(LDrawModel*);
		virtual void readComment(char*);
		virtual bool readSubModel(char*);
		virtual void readLine(char*);
		virtual void readTriangle(char*);
		virtual void readQuad(char*);
		virtual int colinearIndex(const Vector& p0, const Vector& p1,
			const Vector& p2, const Vector& p3);
		virtual int colinearIndex(const Vector& p0, const Vector& p1,
			const Vector& p2);
		virtual int middleIndex(const Vector& p0, const Vector& p1,
			const Vector& p2);
		virtual void newQuad(int, Vector&, Vector&, Vector&, Vector&);
		virtual bool splitConcavePoints(int, const Vector&, const Vector&,
			const Vector&, const Vector&);
		virtual void newTriangle(int, const Vector&, const Vector&,
			const Vector&, int *returnIndex = NULL);
		virtual void newLine(int, const Vector&, const Vector&);
		virtual void readConditionalLine(char*);
		virtual void newConditionalLine(int, const Vector&, const Vector&, 
			const Vector&, const Vector&);

//		virtual void addShape(TCObjectArray*&, TGLShape*);
//		virtual void reallocateShapes(TGLShape**&, int, int);
		virtual void addQuadStrip(TGLShape*);
		virtual void addQuad(TGLShape*);
		virtual void addTriangle(TGLShape*);
		virtual void addLine(TGLShape*);
		virtual void addLineStrip(TGLShape*);
//		virtual void addSubModel(LDrawModel*);

//		virtual void setGlColor(int);
		virtual FILE* openSubModelNamed(char*, char*);

		virtual bool readFile(bool);
		virtual bool parseLines(void);
		virtual void addFileLine(char*);
		virtual TCDictionary* getLoadedModels(void);
		virtual LDrawModel* subModelNamed(char*, bool = true, bool = true);
		virtual void drawShapes(TGLShapeArray*, float*, int, int);
		virtual int swapPointsIfNeeded(Vector&, Vector&, Vector&, Vector&);
		virtual void swapPoints(Vector&, Vector&);
		virtual int swapNeeded(const Vector&, const Vector&, const Vector&,
			const Vector&);
//		virtual void computeChildren(float*);
//		virtual void compute(float*);
//		virtual void computeShapes(TCObjectArray*, float*);
		virtual void flatten(LDrawModel*, float*, int = 16);
		virtual void flattenShape(TGLShape*, float*, int);
		virtual void countSubShapes(LDrawModel* subModel, int* counts);
		virtual void countSubShapes(LDrawModel*, int&, int&, int&, int&, int&);
		virtual bool initializeNewSubModel(LDrawModel*, char*, FILE*,
			bool = true, bool = true);
		virtual void shrink(void);
		virtual int getHighlightColor(int);
		virtual void enableLighting(void);
		virtual void disableLighting(void);
		virtual void clearModel(void);
		virtual void convertStud(int numSegments);
		virtual void convertStu2(void);
		virtual void convertStu22(bool = 0);
		virtual void convertStu23(bool = 0);
		virtual void convertStu24(bool = 0);
		virtual void convertCylinder(float = 1.0f);
		virtual void convertCyls(float = 1.0f);
		virtual void convertCyls2(float = 1.0f);
		virtual void convertCone(float = 1.0f, int = 1);
		virtual void convertRing(float = 1.0f, int = 1);
		virtual void convertDisk(float = 1.0f);
		virtual void convertNdis(float = 1.0f);
		virtual void convertEdge(float = 1.0f);
		virtual void convertCcyli(float = 1.0f);
		virtual void convertEighthSphere(void);
		virtual void convertTorusO(int numSegments, float size);
		virtual void convertTorusI(int numSegments, float size);
//		virtual void deleteSubModels(void);
		virtual int numCylinderSegments(void);
		virtual bool startsWithFraction(char *filename);
		virtual float startingFraction(char *filename);
		virtual bool isCyli(char *filename);
		virtual bool isCyls(char *filename);
		virtual bool isCyls2(char *filename);
		virtual bool isDisc(char *filename);
		virtual bool isNdis(char *filename);
		virtual bool isEdge(char *filename);
		virtual bool isCcyli(char *filename);
		virtual bool isCon(char *filename);
		virtual bool is1DigitCon(char *filename);
		virtual bool is2DigitCon(char *filename);
		virtual bool isRing(char *filename);
		virtual bool isFracRing(char *filename);
		virtual bool isFracRin(char *filename);
		virtual bool isTorusO(char *filename);
		virtual bool isTorusI(char *filename);
		void addCylinder(const Vector&, float, float, int, int = -1);
		void addCyls(const Vector&, float, float, int, int = -1);
		void addCyls2(const Vector&, float, float, int, int = -1);
		void addCcyli(const Vector&, float, float, int, int = -1);
		void addTorusO(const Vector& center, float yRadius, float xzRadius,
			int numSegments, int usedSegments);
		void addTorusI(const Vector& center, float yRadius, float xzRadius,
			int numSegments, int usedSegments);
		void addCone(const Vector&, float, float, float, int, int = -1);
		void addDisk(const Vector&, float, int, int = -1);
		void addNotDisk(const Vector&, float, int, int = -1);
		void addRing(const Vector&, float, float, int, int = -1);
		void addEdge(const Vector&, float, int, int = -1);
		void addEighthSphere(const Vector&, float, int);
		void genConeConditionals(TGLShape *cone, int numSegments,
			int usedSegments);
		void genCylsConditionals(TGLShape *cone, int numSegments,
			int usedSegments);
		void genCyls2Conditionals(TGLShape *cone, int numSegments,
			int usedSegments);
		void genEighthSphereConditionals(TGLEighthSphere *eighthSphere,
			int numSegments);
		void genTorusConditionals(TGLTorus *torus, int numSegments,
			int usedSegments);
		void genTorusOConditionals(TGLTorusQ *torus, int usedSegments,
			const Vector &center, float height);
		void genTorusIConditionals(TGLTorusQ *torus, int numSegments,
			int usedSegments, const Vector &center, float radius, float height);
		virtual void setCirclePoint(float, float, const Vector&, Vector&);
		virtual void performPrimitiveSubstitution(void);
		virtual Vector rearrangeVector(const Vector&, int, int, int);
		virtual Vector findLinePoint(const Vector&);
		virtual int checkLinePoint(const Vector&, const Vector&);
		virtual Vector calcIntersection(int, int, int, Vector*, Vector*,
			Vector*);
		virtual bool isPartOrSubpart(void);
		void commonInit(void);
		void initShapes(void);
		virtual TGLShape* getShape(int type, int offset);
		virtual TGLShapeArray* getShapes(int type);
		virtual void addShape(int type, TGLShape* shape);
		virtual int reportProgress(char* message, float progress);
		virtual int reportError(LDMError* error);
		virtual int reportError(char* format, LDMErrorType type, ...);
		virtual int reportQuadSplit(const Vector& q1, const Vector& q2,
			const Vector& q3, const Vector& q4, const Vector& t1,
			const Vector& t2, const Vector& t3, const Vector& t4,
			const Vector& t5, const Vector& t6);
//		virtual int reportError(char* message, LDMErrorType type);
		virtual void propagateColors(void);
		virtual void propagateColors(float* matrix, int colorNumber);
		virtual void addTransparentShape(int type, TGLShape* shape);
		virtual void sortTransparentShapes(void);
		virtual void prepareForTransparentSort(float* matrix);
		bool getUseStipple(void)
		{
			return topParent ? topParent->flags.useStipple : flags.useStipple;
		}
		bool shouldSortTransparent(void)
		{
			return topParent && topParent != this ?
				topParent->shouldSortTransparent() :
				flags.sortTransparent && !flags.useStipple;
		}
		virtual int getColorNumberForRGB(TCByte r, TCByte g,
			TCByte b);
		virtual bool isColorNumberRGB(int colorNumber, TCByte r,
			TCByte g, TCByte b);
		virtual void finishPart(bool allowScale = true);
		virtual void smooth(bool useTree);
		virtual bool shouldLoadConditionalLines(void);
		virtual bool shouldFlipNormal(const Vector &normal1,
			const Vector &normal2);

		virtual void fillConditionalDict(TGLShapeArrayDictionary
			*conditionalDict);
		virtual void addConditionalPoint(TGLShapeArrayDictionary
			*conditionalDict, TGLShape *line, int pointIndex);
		virtual void calcShapeNormals(TGLShapeArrayDictionary *conditionalDict,
			int shapeIndex);
		virtual void processSmoothEdge(TGLShapeArrayDictionary *conditionalDict,
			TGLShape *shape, int index0, int index1, const Vector &normal);
		virtual void normalizeNormals(TGLShapeArrayDictionary *conditionalDict);
		virtual void applyShapeNormals(TGLShapeArrayDictionary *conditionalDict,
			int shapeIndex);
		virtual void applySmoothEdge(TGLShapeArrayDictionary *conditionalDict,
			TGLShape *shape, int index0, int index1, const Vector &normal);

		virtual void addConditionalPoint(VectorShapeArrayTree *conditionalTree,
			TGLShape *line, int pointIndex);
		virtual void fillConditionalTree(VectorShapeArrayTree *conditionalTree);
		virtual void processSmoothEdge(VectorShapeArrayTree *conditionalTree,
			TGLShape *shape, int index0, int index1, const Vector &normal);
		virtual void calcShapeNormals(VectorShapeArrayTree *conditionalTree,
			int shapeIndex);
		virtual void normalizeNormals(VectorShapeArrayTree *conditionalTree);
		virtual void applySmoothEdge(VectorShapeArrayTree *conditionalTree,
			TGLShape *shape, int index0, int index1, const Vector &normal);
		virtual void applyShapeNormals(VectorShapeArrayTree *conditionalTree,
			int shapeIndex);
		virtual void applySpecialColorSettings(int colorNumber, int color);

		static void normalizeFunc(const void *key, TCObject *value);

		static int shapeSortFunction(const void* left, const void* right);
		static bool colorNumberIsTransparent(int);

		LDrawModel* parent;
		LDrawModel* topParent;
		TCObjectArrayArray* shapes;
		LDrawSubModelArray* newSubModels;
//		LDrawModel** subModels;
//		float** transformationMatrices;
//		int* colors;
//		int* colorNumbers;
//		int subModelCount;
//		int subModelsAllocated;
		LDMFileLineArray* fileLines;
		char* name;
		char* filename;
		GLuint listId;
		LDMProgressCallback progressCallback;
		void* progressUserData;
		LDMErrorCallback errorCallback;
		void* errorUserData;
		LDMFileLine* currentFileLine;
		int curveQuality;
		int defaultColorNumber;
		struct
		{
			int cullBackFaces:1;
			int compiled:1;
			bool allowLowResStuds:1;
			int drawHighlights:1;
			int subModelIsPart:1;
			int haveMainModel:1;
			int isPart:1;
			int isForcedPart:1;
			int normalsComputed:1;
			int allowPrimitiveSubstitution:1;
			bool useStipple:1;
			int readCanceled:1;
			int sortTransparent:1;
			bool blackHighlights:1;
			bool cutawayDraw:1;
			bool drawConditionalHighlights:1;
			bool performSmoothing:1;
			bool lineSmoothing:1;
			bool compiling:1;
			int _reserved:14;
		} flags;

//		static int lastColor;
		static TCDictionary* loadedModels;
		static float seamWidth;
		static int usePolygonOffset;
		static int useLighting;
		static int drawWireframe;
		static char* systemLDrawDir;
		static GLPolygonOffsetFunc polygonOffsetFunc;
		static LDMSpecialColorSettings specialColors[NUM_SPECIAL_COLORS];

		static class LDrawModelCleanup
		{
		public:
			~LDrawModelCleanup(void);
		} modelCleanup;
		friend LDrawModelCleanup;
};

#endif // __LDRAWMODEL_H__
