#include <shlobj.h>
#include <shlwapi.h>
#include "LDViewPreferences.h"
#include <LDLib/LDrawModelViewer.h>
#include <LDLib/TGLShape.h>
#include "LDVExtensionsSetup.h"
#include "AppResources.h"
#include "UserDefaultsKeys.h"

#include <TCFoundation/TCUserDefaults.h>
#include <TCFoundation/mystring.h>
#include <Commctrl.h>
#include <stdio.h>

#ifndef IDC_HARDWARE_STEREO
#define IDC_HARDWARE_STEREO 1030
#endif

LDViewPreferences::LDViewPreferences(HINSTANCE hInstance,
									 LDrawModelViewer* modelViewer)
	:CUIPropertySheet("LDView Preferences", hInstance),
	modelViewer(modelViewer ? ((LDrawModelViewer*)modelViewer->retain()) : NULL),
	generalPageNumber(0),
	geometryPageNumber(1),
	effectsPageNumber(2),
	primitivesPageNumber(3),
	hGeneralPage(NULL),
	hButtonColorDC(NULL),
	hGeometryPage(NULL),
	hEffectsPage(NULL),
	hPrimitivesPage(NULL)
{
	loadSettings();
	applySettings();
}

LDViewPreferences::~LDViewPreferences(void)
{
}

void LDViewPreferences::dealloc(void)
{
	if (modelViewer)
	{
		modelViewer->release();
	}
	CUIPropertySheet::dealloc();
}

void LDViewPreferences::applySettings(void)
{
	applyGeneralSettings();
	applyGeometrySettings();
	applyEffectsSettings();
	applyPrimitivesSettings();
	if (modelViewer)
	{
		modelViewer->setZoomMax(zoomMax);
		modelViewer->setShowAllConditionalLines(showAllConditionalLines);
		modelViewer->setShowConditionalControlPoints(
			showConditionalControlPoints);
	}
}

void LDViewPreferences::applyGeneralSettings(void)
{
	if (modelViewer)
	{
		int r, g, b;

		// FSAA taken care of automatically.
		// showFrameRate taken care of automatically.
		// showErrors taken care of automatically.
		// fullScreenRefresh taken care of automatically.
		getRGB(backgroundColor, r, g, b);
		modelViewer->setBackgroundRGBA(r, g, b, 0);
		getRGB(defaultColor, r, g, b);
		modelViewer->setDefaultRGB((BYTE)r, (BYTE)g, (BYTE)b);
		modelViewer->setDefaultColorNumber(defaultColorNumber);
		modelViewer->setLineSmoothing(lineSmoothing);
	}
}

void LDViewPreferences::applyGeometrySettings(void)
{
	if (modelViewer)
	{
		if (useSeams)
		{
			modelViewer->setSeamWidth(seamWidth / 100.0f);
		}
		else
		{
			modelViewer->setSeamWidth(0.0f);
		}
		modelViewer->setDrawWireframe(drawWireframe);
		modelViewer->setUseWireframeFog(useWireframeFog);
		modelViewer->setWireframeLineWidth((GLfloat)wireframeThickness);
		modelViewer->setShowsHighlightLines(showsHighlightLines);
		modelViewer->setDrawConditionalHighlights(drawConditionalHighlights);
		modelViewer->setUsePolygonOffset(usePolygonOffset);
		modelViewer->setBlackHighlights(blackHighlights);
		modelViewer->setHighlightLineWidth((GLfloat)edgeThickness);
	}
}

void LDViewPreferences::applyEffectsSettings(void)
{
	if (modelViewer)
	{
		modelViewer->setUseLighting(useLighting);
		modelViewer->setQualityLighting(qualityLighting);
		modelViewer->setSubduedLighting(subduedLighting);
		modelViewer->setUsesSpecular(usesSpecular);
		modelViewer->setOneLight(oneLight);
		modelViewer->setStereoMode(stereoMode);
		modelViewer->setStereoEyeSpacing((GLfloat)stereoEyeSpacing);
		modelViewer->setCutawayMode(cutawayMode);
		modelViewer->setCutawayAlpha((float)cutawayAlpha / 100.0f);
		modelViewer->setCutawayLineWidth((float)cutawayThickness);
		modelViewer->setSortTransparent(sortTransparent);
		modelViewer->setUseStipple(useStipple);
		modelViewer->setUsesFlatShading(usesFlatShading);
		modelViewer->setPerformSmoothing(performSmoothing);
	}
}

void LDViewPreferences::applyPrimitivesSettings(void)
{
	if (modelViewer)
	{
		modelViewer->setAllowPrimitiveSubstitution(allowPrimitiveSubstitution);
		modelViewer->setTextureStuds(textureStuds);
		modelViewer->setTextureFilterType(textureFilterType);
		modelViewer->setCurveQuality(curveQuality);
		modelViewer->setQualityStuds(qualityStuds);
	}
}

void LDViewPreferences::loadSettings(void)
{
	loadGeneralSettings();
	loadGeometrySettings();
	loadEffectsSettings();
	loadPrimitivesSettings();

	zoomMax = TCUserDefaults::longForKey(ZOOM_MAX_KEY, 199) / 100.0f;
	showAllConditionalLines = TCUserDefaults::longForKey(SHOW_ALL_TYPE5_KEY)
		!= 0;
	showConditionalControlPoints =
		TCUserDefaults::longForKey(SHOW_TYPE5_CONTROL_POINTS_KEY) != 0;
}

void LDViewPreferences::loadDefaultGeneralSettings(void)
{
	int i;

	fsaaMode = 0;
	lineSmoothing = false;
	showsFPS = false;
	showErrors = true;
	fullScreenRefresh = 0;
	backgroundColor = (COLORREF)0;
	defaultColor = (COLORREF)0x999999;
	defaultColorNumber = -1;
	for (i = 0; i < 16; i++)
	{
		int r, g, b, a;

		LDrawModel::getRGBA(i, r, g, b, a);
		customColors[i] = TGLShape::colorForRGBA(r, g, b, a);
	}
}

void LDViewPreferences::loadDefaultGeometrySettings(void)
{
	useSeams = -1;
	seamWidth = 50;
	drawWireframe = false;
	useWireframeFog = false;
	wireframeThickness = 1;
	showsHighlightLines = false;
	drawConditionalHighlights = false;
	usePolygonOffset = true;
	blackHighlights = false;
	edgeThickness = 1;
}

void LDViewPreferences::loadDefaultEffectsSettings(void)
{
	useLighting = true;
	qualityLighting = false;
	subduedLighting = false;
	usesSpecular = true;
	oneLight = false;
	stereoMode = LDVStereoNone;
	stereoEyeSpacing = 50;
	cutawayMode = LDVCutawayNormal;
	cutawayAlpha = 100;
	cutawayThickness = 1;
	sortTransparent = true;
	useStipple = false;
	usesFlatShading = false;
	performSmoothing = true;
}

void LDViewPreferences::loadDefaultPrimitivesSettings(void)
{
	allowPrimitiveSubstitution = true;
	textureStuds = true;
	textureFilterType = GL_LINEAR_MIPMAP_LINEAR;
	curveQuality = 2;
	qualityStuds = false;
}

void LDViewPreferences::loadGeneralSettings(void)
{
	int i;

	loadDefaultGeneralSettings();
	fsaaMode = TCUserDefaults::longForKey(FSAA_MODE_KEY, (long)fsaaMode);
	lineSmoothing = TCUserDefaults::longForKey(LINE_SMOOTHING_KEY,
		lineSmoothing) != 0;
	showsFPS = TCUserDefaults::longForKey(SHOW_FPS_KEY, (long)showsFPS) != 0;
	showErrors = TCUserDefaults::longForKey(SHOW_ERRORS_KEY, (long)showErrors)
		!= 0;
	fullScreenRefresh = TCUserDefaults::longForKey(FULLSCREEN_REFRESH_KEY,
		fullScreenRefresh);
	backgroundColor =
		(COLORREF)TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY,
		(long)backgroundColor);
	defaultColor = (COLORREF)TCUserDefaults::longForKey(DEFAULT_COLOR_KEY,
		(long)defaultColor);
	defaultColorNumber = TCUserDefaults::longForKey(DEFAULT_COLOR_NUMBER_KEY,
		defaultColorNumber);
	for (i = 0; i < 16; i++)
	{
		char key[128];

		sprintf(key, "%s/Color%02d", CUSTOM_COLORS_KEY, i);
		customColors[i] = TCUserDefaults::longForKey(key, customColors[i]);
	}
}

void LDViewPreferences::loadGeometrySettings(void)
{
	loadDefaultGeometrySettings();
	useSeams = TCUserDefaults::longForKey(SEAMS_KEY, useSeams);
	seamWidth = TCUserDefaults::longForKey(SEAM_WIDTH_KEY, seamWidth);
	if (useSeams == -1)
	{
		if (seamWidth)
		{
			useSeams = 1;
		}
		else
		{
			useSeams = 0;
		}
	}
	drawWireframe = TCUserDefaults::longForKey(WIREFRAME_KEY,
		(long)drawWireframe) != 0;
	useWireframeFog = TCUserDefaults::longForKey(WIREFRAME_FOG_KEY,
		(long)useWireframeFog) != 0;
	wireframeThickness = TCUserDefaults::longForKey(WIREFRAME_THICKNESS_KEY,
		wireframeThickness);
	showsHighlightLines = TCUserDefaults::longForKey(SHOWS_HIGHLIGHT_LINES_KEY,
		(long)showsHighlightLines) != 0;
	drawConditionalHighlights =
		TCUserDefaults::longForKey(CONDITIONAL_HIGHLIGHTS_KEY,
		(long)drawConditionalHighlights) != 0;
	usePolygonOffset = TCUserDefaults::longForKey(POLYGON_OFFSET_KEY,
		(long)usePolygonOffset) != 0;
	blackHighlights = TCUserDefaults::longForKey(BLACK_HIGHLIGHTS_KEY,
		(long)blackHighlights) != 0;
	edgeThickness = TCUserDefaults::longForKey(EDGE_THICKNESS_KEY,
		edgeThickness);
}

void LDViewPreferences::loadEffectsSettings(void)
{
	loadDefaultEffectsSettings();
	useLighting = TCUserDefaults::longForKey(LIGHTING_KEY, (long)useLighting)
		!= 0;
	qualityLighting = TCUserDefaults::longForKey(QUALITY_LIGHTING_KEY,
		(long)qualityLighting) != 0;
	subduedLighting = TCUserDefaults::longForKey(SUBDUED_LIGHTING_KEY,
		(long)subduedLighting) != 0;
	usesSpecular = TCUserDefaults::longForKey(SPECULAR_KEY, (long)usesSpecular)
		!= 0;
	oneLight = TCUserDefaults::longForKey(ONE_LIGHT_KEY, (long)oneLight) != 0;
	stereoMode = (LDVStereoMode)TCUserDefaults::longForKey(STEREO_MODE_KEY,
		(long)stereoMode);
	stereoEyeSpacing = TCUserDefaults::longForKey(STEREO_SPACING_KEY,
		stereoEyeSpacing);
	cutawayMode = (LDVCutawayMode)TCUserDefaults::longForKey(CUTAWAY_MODE_KEY,
		(long)cutawayMode);
	cutawayAlpha = TCUserDefaults::longForKey(CUTAWAY_ALPHA_KEY, cutawayAlpha);
	cutawayThickness = TCUserDefaults::longForKey(CUTAWAY_THICKNESS_KEY,
		cutawayThickness);
	sortTransparent = TCUserDefaults::longForKey(SORT_KEY,
		(long)sortTransparent) != 0;
	useStipple = TCUserDefaults::longForKey(STIPPLE_KEY, (long)useStipple) != 0;
	usesFlatShading = TCUserDefaults::longForKey(FLAT_SHADING_KEY,
		(long)usesFlatShading) != 0;
	performSmoothing = TCUserDefaults::longForKey(PERFORM_SMOOTHING_KEY,
		(long)performSmoothing) != 0;
}

void LDViewPreferences::loadPrimitivesSettings(void)
{
	loadDefaultPrimitivesSettings();
	allowPrimitiveSubstitution = TCUserDefaults::longForKey(
		PRIMITIVE_SUBSTITUTION_KEY, (long)allowPrimitiveSubstitution) != 0;
	textureStuds = TCUserDefaults::longForKey(TEXTURE_STUDS_KEY,
		(long)textureStuds) != 0;
	textureFilterType = TCUserDefaults::longForKey(TEXTURE_FILTER_TYPE_KEY,
		textureFilterType);
	curveQuality = TCUserDefaults::longForKey(CURVE_QUALITY_KEY, curveQuality);
	qualityStuds = TCUserDefaults::longForKey(QUALITY_STUDS_KEY,
		(long)qualityStuds) != 0;
}

int LDViewPreferences::run(void)
{
	addPage(IDD_GENERAL_PREFS);
	addPage(IDD_GEOMETRY_PREFS); 
	addPage(IDD_EFFECTS_PREFS);
	addPage(IDD_PRIMITIVES_PREFS);
	if (modelViewer)
	{
		modelViewer->pause();
	}
	return CUIPropertySheet::run();
}

BOOL LDViewPreferences::doDialogNotify(HWND hDlg, int controlId,
									   LPNMHDR notification)
{
//	printf("LDViewPreferences::doDialogNotify: %d 0x%08X\n", notification->code,
//		notification->code);
	if (notification->code == NM_RELEASEDCAPTURE)
	{
		if (hDlg == hEffectsPage)
		{
			if (controlId == IDC_STEREO_SPACING ||
				controlId == IDC_CUTAWAY_OPACITY ||
				controlId == IDC_CUTAWAY_THICKNESS)
			{
				enableApply(hEffectsPage);
				return FALSE;
			}
		}
		else if (hDlg == hGeometryPage)
		{
			if (controlId == IDC_WIREFRAME_THICKNESS ||
				controlId == IDC_EDGE_THICKNESS)
			{
				enableApply(hGeometryPage);
				return FALSE;
			}
		}
		else if (hDlg == hPrimitivesPage)
		{
			if (controlId == IDC_CURVE_QUALITY)
			{
				enableApply(hPrimitivesPage);
				return FALSE;
			}
		}
	}
	return CUIPropertySheet::doDialogNotify(hDlg, controlId, notification);
}

char* LDViewPreferences::getLDViewPath(bool useQuotes)
{
	return getLDViewPath(NULL, useQuotes);
}

char* LDViewPreferences::getLDViewPath(char* helpFilename, bool useQuotes)
{
	char* programPath = copyString(GetCommandLine());
	char tmpPath[MAX_PATH];

	PathRemoveArgs(programPath);
	PathRemoveFileSpec(programPath);
	strcpy(tmpPath, programPath);
	if (helpFilename)
	{
		PathAppend(tmpPath, helpFilename);
	}
	if (tmpPath[0] == '"')
	{
		if (useQuotes)
		{
			strcat(tmpPath, "\"");
		}
		else
		{
			memmove(tmpPath, tmpPath + 1, strlen(tmpPath));
		}
	}
	delete programPath;
	return copyString(tmpPath);
}

BOOL LDViewPreferences::doDialogHelp(HWND hDlg, LPHELPINFO helpInfo)
{
	DWORD dialogId = 0;
	DWORD helpId;
	char* helpPath = getLDViewPath("LDView.hlp");
	BOOL retValue = FALSE;

	if (hDlg == hGeneralPage)
	{
		dialogId = IDD_GENERAL_PREFS;
	}
	else if (hDlg == hGeometryPage)
	{
		dialogId = IDD_GEOMETRY_PREFS;
	}
	else if (hDlg == hEffectsPage)
	{
		dialogId = IDD_EFFECTS_PREFS;
	}
	else if (hDlg == hPrimitivesPage)
	{
		dialogId = IDD_PRIMITIVES_PREFS;
	}
	if (dialogId)
	{
		helpId = 0x80000000 | (dialogId << 16) | (DWORD)helpInfo->iCtrlId;
		WinHelp((HWND)helpInfo->hItemHandle, helpPath, HELP_CONTEXTPOPUP,
			helpId);
		retValue = FALSE;
	}
	delete helpPath;
	return retValue;
}

BOOL LDViewPreferences::doDialogCommand(HWND hDlg, int controlId,
										int notifyCode, HWND controlHWnd)
{
	if (notifyCode == BN_CLICKED)
	{
		if (hDlg == hGeneralPage || hDlg == hGeometryPage ||
			hDlg == hEffectsPage || hDlg == hPrimitivesPage)
		{
			return doClick(hDlg, controlId, controlHWnd);
		}
	}
	else if (notifyCode == EN_CHANGE)
	{
		if (controlId == IDC_FS_RATE)
		{
			enableApply(hDlg);
			return 0;
		}
	}
	return LDP_UNKNOWN_COMMAND;
}

void LDViewPreferences::setupBackgroundColorButton(void)
{
	setupColorButton(hGeneralPage, hBackgroundColorButton,
		IDC_BACKGROUND_COLOR, hBackgroundColorBitmap, backgroundColor);
}

void LDViewPreferences::setupDefaultColorButton(void)
{
	setupColorButton(hGeneralPage, hDefaultColorButton,
		IDC_DEFAULT_COLOR, hDefaultColorBitmap, defaultColor);
}

void LDViewPreferences::setupColorButton(HWND hPage, HWND &hColorButton,
										 int controlID, HBITMAP &hButtonBitmap,
										 COLORREF color)
{
	RECT clientRect;
	int imageWidth;
	int imageHeight;
	HDC hdc;

	hColorButton = GetDlgItem(hPage, controlID);
	GetClientRect(hColorButton, &clientRect);
	imageWidth = clientRect.right - clientRect.left - 10;
	imageHeight = clientRect.bottom - clientRect.top - 10;
	hdc = GetDC(hPage);
	if (!hButtonColorDC)
	{
		hButtonColorDC = CreateCompatibleDC(hdc);
	}
	hButtonBitmap = CreateCompatibleBitmap(hdc, imageWidth, imageHeight);
	ReleaseDC(hPage, hdc);
	SetBitmapDimensionEx(hButtonBitmap, imageWidth, imageHeight, NULL);
	redrawColorBitmap(hColorButton, hButtonBitmap, color);
	SendDlgItemMessage(hPage, controlID, BM_SETIMAGE, (WPARAM)IMAGE_BITMAP,
		(LPARAM)hButtonBitmap);
}

void LDViewPreferences::enableSeams(void)
{
	char seamWidthString[128];

	EnableWindow(hSeamSpin, TRUE);
	SendDlgItemMessage(hGeometryPage, IDC_SEAMS, BM_SETCHECK, 1, 0);
	sprintf(seamWidthString, "%0.2f", seamWidth / 100.0f);
	SendDlgItemMessage(hGeometryPage, IDC_SEAM_WIDTH_FIELD, WM_SETTEXT, 0,
		(LPARAM)seamWidthString);
}

void LDViewPreferences::disableSeams(void)
{
	EnableWindow(hSeamSpin, FALSE);
	SendDlgItemMessage(hGeometryPage, IDC_SEAMS, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeometryPage, IDC_SEAM_WIDTH_FIELD, WM_SETTEXT, 0,
		(LPARAM)"");
}

void LDViewPreferences::enableTextureFiltering(void)
{
	int activeTextureFilter;

	EnableWindow(hTextureNearestButton, TRUE);
	EnableWindow(hTextureBilinearButton, TRUE);
	EnableWindow(hTextureTrilinearButton, TRUE);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_NEAREST, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_BILINEAR, BM_SETCHECK, 0,
		0);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_TRILINEAR, BM_SETCHECK, 0,
		0);
	switch (textureFilterType)
	{
	case GL_NEAREST_MIPMAP_NEAREST:
		activeTextureFilter = IDC_TEXTURE_NEAREST;
		break;
	case GL_LINEAR_MIPMAP_NEAREST:
		activeTextureFilter = IDC_TEXTURE_BILINEAR;
		break;
	case GL_LINEAR_MIPMAP_LINEAR:
		activeTextureFilter = IDC_TEXTURE_TRILINEAR;
		break;
	default:
		textureFilterType = GL_LINEAR_MIPMAP_LINEAR;
		activeTextureFilter = IDC_TEXTURE_TRILINEAR;
		break;
	}
	SendDlgItemMessage(hPrimitivesPage, activeTextureFilter, BM_SETCHECK, 1, 0);
}

void LDViewPreferences::disableTextureFiltering(void)
{
	EnableWindow(hTextureNearestButton, FALSE);
	EnableWindow(hTextureBilinearButton, FALSE);
	EnableWindow(hTextureTrilinearButton, FALSE);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_NEAREST, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_BILINEAR, BM_SETCHECK, 0,
		0);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_TRILINEAR, BM_SETCHECK, 0,
		0);
}

void LDViewPreferences::enablePrimitives(void)
{
	EnableWindow(hTextureStudsButton, TRUE);
	EnableWindow(hCurveQualityLabel, TRUE);
	EnableWindow(hCurveQualitySlider, TRUE);
	SendDlgItemMessage(hPrimitivesPage, IDC_PRIMITIVE_SUBSTITUTION, BM_SETCHECK,
		1, 0);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_STUDS, BM_SETCHECK,
		textureStuds, 0);
	if (textureStuds)
	{
		enableTextureFiltering();
	}
	else
	{
		disableTextureFiltering();
	}
}

void LDViewPreferences::disablePrimitives(void)
{
	EnableWindow(hTextureStudsButton, FALSE);
	EnableWindow(hCurveQualityLabel, FALSE);
	EnableWindow(hCurveQualitySlider, FALSE);
	SendDlgItemMessage(hPrimitivesPage, IDC_PRIMITIVE_SUBSTITUTION, BM_SETCHECK,
		0, 0);
	SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_STUDS, BM_SETCHECK, 0, 0);
	disableTextureFiltering();
}

void LDViewPreferences::setupSeamWidth(void)
{
	UDACCEL accels[] = {{0, 1}, {1, 10}, {2, 20}};

	hSeamSpin = GetDlgItem(hGeometryPage, IDC_SEAM_SPIN);
	SendDlgItemMessage(hGeometryPage, IDC_SEAM_SPIN, UDM_SETRANGE, 0,
		MAKELONG((short)500, (short)0)); 
	SendDlgItemMessage(hGeometryPage, IDC_SEAM_SPIN, UDM_SETPOS, 0, seamWidth);
	SendDlgItemMessage(hGeometryPage, IDC_SEAM_SPIN, UDM_SETACCEL,
		sizeof(accels) / sizeof(UDACCEL), (long)accels);
	if (useSeams)
	{
		enableSeams();
	}
	else
	{
		disableSeams();
	}
}

void LDViewPreferences::setupFullScreenRefresh(void)
{
	hFullScreenRateField = GetDlgItem(hGeneralPage, IDC_FS_RATE);
	if (fullScreenRefresh)
	{
		char buf[128];

		sprintf(buf, "%d", fullScreenRefresh);
		buf[3] = 0;
		SendDlgItemMessage(hGeneralPage, IDC_FS_RATE, WM_SETTEXT, 0,
			(LPARAM)buf);
		SendDlgItemMessage(hGeneralPage, IDC_FS_REFRESH, BM_SETCHECK, 1, 0);
		EnableWindow(hFullScreenRateField, TRUE);
	}
	else
	{
		SendDlgItemMessage(hGeneralPage, IDC_FS_RATE, WM_SETTEXT, 0,
			(LPARAM)"");
		SendDlgItemMessage(hGeneralPage, IDC_FS_REFRESH, BM_SETCHECK, 0, 0);
		EnableWindow(hFullScreenRateField, FALSE);
	}
}

/*
void LDViewPreferences::drawButtonBorder(HDC hdc, COLORREF color1,
										 COLORREF color2, RECT rect)
{
	HPEN hPen1 = CreatePen(PS_SOLID, 1, color1);
	HPEN hPen2 = CreatePen(PS_SOLID, 1, color2);
	HPEN hOldPen;

	hOldPen = (HPEN)SelectObject(hdc, hPen1);
	MoveToEx(hBackgroundColorDC, rect.left, rect.bottom, NULL);
	LineTo(hBackgroundColorDC, rect.left, rect.top);
	LineTo(hBackgroundColorDC, rect.right, rect.top);
	SelectObject(hBackgroundColorDC, hPen2);
	LineTo(hBackgroundColorDC, rect.right, rect.bottom);
	LineTo(hBackgroundColorDC, rect.left, rect.bottom);
	MoveToEx(hBackgroundColorDC, rect.left, rect.bottom, NULL);
	LineTo(hBackgroundColorDC, rect.right, rect.bottom);
	SelectObject(hdc, hOldPen);
	DeleteObject(hPen1);
	DeleteObject(hPen2);
}
*/

void LDViewPreferences::redrawColorBitmap(HWND hColorButton,
										  HBITMAP hButtonBitmap, COLORREF color)
{
	HBRUSH hBrush = CreateSolidBrush(color);
	RECT bitmapRect;
	SIZE bitmapSize;
	HBITMAP hOldBitmap;
	HBRUSH hOldBrush;

	GetBitmapDimensionEx(hButtonBitmap, &bitmapSize);
	bitmapRect.left = 0;
	bitmapRect.top = 0;
	bitmapRect.right = bitmapSize.cx - 1;
	bitmapRect.bottom = bitmapSize.cy - 1;
	hOldBrush = (HBRUSH)SelectObject(hButtonColorDC, hBrush);
	hOldBitmap = (HBITMAP)SelectObject(hButtonColorDC, hButtonBitmap);
	FillRect(hButtonColorDC, &bitmapRect, hBrush);
	bitmapRect.top++;
	bitmapRect.left++;
	InflateRect(&bitmapRect, 1, 1);
	DrawEdge(hButtonColorDC, &bitmapRect, EDGE_SUNKEN,
		BF_BOTTOMLEFT | BF_TOPRIGHT);
	SelectObject(hButtonColorDC, hOldBrush);
	SelectObject(hButtonColorDC, hOldBitmap);
	DeleteObject(hBrush);
	InvalidateRect(hColorButton, NULL, 1);
}

void LDViewPreferences::applyGeneralChanges(void)
{
	if (hGeneralPage)
	{
		int temp;
		bool bTemp;
		int i;

		if (TCUserDefaults::longForKey(FSAA_MODE_KEY) != fsaaMode)
		{
			TCUserDefaults::setLongForKey(fsaaMode, FSAA_MODE_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hGeneralPage, IDC_LINE_AA,
			BM_GETCHECK, 0, 0) != 0) != lineSmoothing)
		{
			lineSmoothing = bTemp;
			TCUserDefaults::setLongForKey(lineSmoothing, LINE_SMOOTHING_KEY);
			if (modelViewer)
			{
				modelViewer->setLineSmoothing(lineSmoothing);
			}
		}
		if ((bTemp = SendDlgItemMessage(hGeneralPage, IDC_FRAME_RATE,
			BM_GETCHECK, 0, 0) != 0) != showsFPS)
		{
			showsFPS = bTemp;
			TCUserDefaults::setLongForKey(showsFPS, SHOW_FPS_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hGeneralPage, IDC_SHOW_ERRORS,
			BM_GETCHECK, 0, 0) != 0) != showErrors)
		{
			showErrors = bTemp;
			TCUserDefaults::setLongForKey(showErrors, SHOW_ERRORS_KEY);
		}
		temp = SendDlgItemMessage(hGeneralPage, IDC_FS_REFRESH, BM_GETCHECK, 0,
			0);
		if (temp)
		{
			char buf[4];

			SendDlgItemMessage(hGeneralPage, IDC_FS_RATE, WM_GETTEXT, 4,
				(LPARAM)buf);
			if (sscanf(buf, "%d", &temp) == 1)
			{
				if (!temp)
				{
					temp = -1;
				}
			}
			else
			{
				temp = -1;
			}
		}
		if (temp != fullScreenRefresh && temp >= 0)
		{
			fullScreenRefresh = temp;
			TCUserDefaults::setLongForKey(fullScreenRefresh,
				FULLSCREEN_REFRESH_KEY);
		}
		if ((COLORREF)TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY) !=
			backgroundColor)
		{
			int r, g, b;

			getRGB(backgroundColor, r, g, b);
			TCUserDefaults::setLongForKey((long)backgroundColor,
				BACKGROUND_COLOR_KEY);
			if (modelViewer)
			{
				modelViewer->setBackgroundRGBA(r, g, b, 0);
			}
		}
		if ((COLORREF)TCUserDefaults::longForKey(DEFAULT_COLOR_KEY) !=
			defaultColor)
		{
			int r, g, b;

			getRGB(defaultColor, r, g, b);
			TCUserDefaults::setLongForKey((long)defaultColor,
				DEFAULT_COLOR_KEY);
			if (modelViewer)
			{
				modelViewer->setDefaultRGB((BYTE)r, (BYTE)g, (BYTE)b);
			}
		}
		for (i = 0; i < 16; i++)
		{
			if (customColors[i])
			{
				char key[128];

				sprintf(key, "%s/Color%02d", CUSTOM_COLORS_KEY, i);
				TCUserDefaults::setLongForKey((long)customColors[i], key);
			}
		}
	}
}

void LDViewPreferences::applyGeometryChanges(void)
{
	if (hGeometryPage)
	{
		int temp;
		bool bTemp;

		if ((temp = SendDlgItemMessage(hGeometryPage, IDC_SEAMS, BM_GETCHECK, 0,
			0)) != useSeams)
		{
			useSeams = temp;
			if (modelViewer)
			{
				if (useSeams)
				{
					modelViewer->setSeamWidth(seamWidth / 100.0f);
				}
				else
				{
					modelViewer->setSeamWidth(0.0f);
				}
			}
			TCUserDefaults::setLongForKey(useSeams, SEAMS_KEY);
		}
		if (useSeams && TCUserDefaults::longForKey(SEAM_WIDTH_KEY) != seamWidth)
		{
			TCUserDefaults::setLongForKey(seamWidth, SEAM_WIDTH_KEY);
			if (useSeams)
			{
				if (modelViewer)
				{
					modelViewer->setSeamWidth(seamWidth / 100.0f);
				}
			}
		}
		else
		{
			seamWidth = TCUserDefaults::longForKey(SEAM_WIDTH_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hGeometryPage, IDC_WIREFRAME,
			BM_GETCHECK, 0, 0) != 0) != drawWireframe)
		{
			drawWireframe = bTemp;
			if (modelViewer)
			{
				modelViewer->setDrawWireframe(drawWireframe);
			}
			TCUserDefaults::setLongForKey(drawWireframe, WIREFRAME_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hGeometryPage, IDC_WIREFRAME_FOG,
			BM_GETCHECK, 0, 0) != 0) != useWireframeFog)
		{
			useWireframeFog = bTemp;
			if (modelViewer)
			{
				modelViewer->setUseWireframeFog(useWireframeFog);
			}
			TCUserDefaults::setLongForKey(useWireframeFog, WIREFRAME_FOG_KEY);
		}
		if ((temp = SendDlgItemMessage(hGeometryPage, IDC_WIREFRAME_THICKNESS,
			TBM_GETPOS, 0, 0)) != wireframeThickness)
		{
			wireframeThickness = temp;
			if (modelViewer)
			{
				modelViewer->setWireframeLineWidth((GLfloat)wireframeThickness);
			}
			TCUserDefaults::setLongForKey(wireframeThickness, 
				WIREFRAME_THICKNESS_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hGeometryPage, IDC_HIGHLIGHTS,
			BM_GETCHECK, 0, 0) != 0) != showsHighlightLines)
		{
			showsHighlightLines = bTemp;
			if (modelViewer)
			{
				modelViewer->setShowsHighlightLines(showsHighlightLines);
			}
			TCUserDefaults::setLongForKey(showsHighlightLines,
				SHOWS_HIGHLIGHT_LINES_KEY);
		}
		if (showsHighlightLines && (bTemp = SendDlgItemMessage(hGeometryPage,
			IDC_CONDITIONAL_HIGHLIGHTS, BM_GETCHECK, 0, 0) != 0) !=
			drawConditionalHighlights)
		{
			drawConditionalHighlights = bTemp;
			if (modelViewer)
			{
				modelViewer->setDrawConditionalHighlights(
					drawConditionalHighlights);
			}
			TCUserDefaults::setLongForKey(drawConditionalHighlights,
				CONDITIONAL_HIGHLIGHTS_KEY);
		}
		if (showsHighlightLines && (bTemp = SendDlgItemMessage(hGeometryPage,
			IDC_QUALITY_LINES, BM_GETCHECK, 0, 0) != 0) !=
			usePolygonOffset)
		{
			usePolygonOffset = bTemp;
			if (modelViewer)
			{
				modelViewer->setUsePolygonOffset(usePolygonOffset);
			}
			TCUserDefaults::setLongForKey(usePolygonOffset, POLYGON_OFFSET_KEY);
		}
		if (showsHighlightLines && (bTemp = SendDlgItemMessage(hGeometryPage,
			IDC_ALWAYS_BLACK, BM_GETCHECK, 0, 0) != 0) !=
			blackHighlights)
		{
			blackHighlights = bTemp;
			if (modelViewer)
			{
				modelViewer->setBlackHighlights(blackHighlights);
			}
			TCUserDefaults::setLongForKey(blackHighlights, 
				BLACK_HIGHLIGHTS_KEY);
		}
		if ((temp = SendDlgItemMessage(hGeometryPage, IDC_EDGE_THICKNESS,
			TBM_GETPOS, 0, 0)) != edgeThickness)
		{
			edgeThickness = temp;
			if (modelViewer)
			{
				modelViewer->setHighlightLineWidth((GLfloat)edgeThickness);
			}
			TCUserDefaults::setLongForKey(edgeThickness,  EDGE_THICKNESS_KEY);
		}
	}
}

void LDViewPreferences::applyEffectsChanges(void)
{
	if (hEffectsPage)
	{
		bool bTemp;
		int temp;

		if ((bTemp = SendDlgItemMessage(hEffectsPage, IDC_LIGHTING,
			BM_GETCHECK, 0, 0) != 0) != useLighting)
		{
			useLighting = bTemp;
			if (modelViewer)
			{
				modelViewer->setUseLighting(useLighting);
			}
			TCUserDefaults::setLongForKey(useLighting, LIGHTING_KEY);
		}
		if (useLighting && (bTemp = SendDlgItemMessage(hEffectsPage,
			IDC_LIGHTING_QUALITY, BM_GETCHECK, 0, 0) != 0) !=
			qualityLighting)
		{
			qualityLighting = bTemp;
			if (modelViewer)
			{
				modelViewer->setQualityLighting(qualityLighting);
			}
			TCUserDefaults::setLongForKey(qualityLighting,
				QUALITY_LIGHTING_KEY);
		}
		if (useLighting && (bTemp = SendDlgItemMessage(hEffectsPage,
			IDC_LIGHTING_SUBDUED, BM_GETCHECK, 0, 0) != 0) !=
			subduedLighting)
		{
			subduedLighting = bTemp;
			if (modelViewer)
			{
				modelViewer->setSubduedLighting(subduedLighting);
			}
			TCUserDefaults::setLongForKey(subduedLighting,
				SUBDUED_LIGHTING_KEY);
		}
		if (useLighting && (bTemp = SendDlgItemMessage(hEffectsPage,
			IDC_SPECULAR, BM_GETCHECK, 0, 0) != 0) != usesSpecular)
		{
			usesSpecular = bTemp;
			if (modelViewer)
			{
				modelViewer->setUsesSpecular(usesSpecular);
			}
			TCUserDefaults::setLongForKey(usesSpecular, SPECULAR_KEY);
		}
		if (useLighting && (bTemp = SendDlgItemMessage(hEffectsPage,
			IDC_ALTERNATE_LIGHTING, BM_GETCHECK, 0, 0) != 0) != oneLight)
		{
			oneLight = bTemp;
			if (modelViewer)
			{
				modelViewer->setOneLight(oneLight);
			}
			TCUserDefaults::setLongForKey(oneLight, ONE_LIGHT_KEY);
		}
		if (TCUserDefaults::longForKey(STEREO_MODE_KEY) != stereoMode)
		{
			if (modelViewer)
			{
				modelViewer->setStereoMode(stereoMode);
			}
			TCUserDefaults::setLongForKey(stereoMode, STEREO_MODE_KEY);
		}
		if ((temp = SendDlgItemMessage(hEffectsPage, IDC_STEREO_SPACING,
			TBM_GETPOS, 0, 0)) != stereoEyeSpacing)
		{
			stereoEyeSpacing = temp;
			if (modelViewer)
			{
				modelViewer->setStereoEyeSpacing((GLfloat)stereoEyeSpacing);
			}
			TCUserDefaults::setLongForKey(stereoEyeSpacing, STEREO_SPACING_KEY);
		}
		if (TCUserDefaults::longForKey(CUTAWAY_MODE_KEY) != cutawayMode)
		{
			if (modelViewer)
			{
				modelViewer->setCutawayMode(cutawayMode);
			}
			TCUserDefaults::setLongForKey(cutawayMode, CUTAWAY_MODE_KEY);
		}
		if ((temp = SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_OPACITY,
			TBM_GETPOS, 0, 0)) != cutawayAlpha)
		{
			cutawayAlpha = temp;
			if (modelViewer)
			{
				modelViewer->setCutawayAlpha((GLfloat)cutawayAlpha / 100.0f);
			}
			TCUserDefaults::setLongForKey(cutawayAlpha, CUTAWAY_ALPHA_KEY);
		}
		if ((temp = SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_THICKNESS,
			TBM_GETPOS, 0, 0)) != cutawayThickness)
		{
			cutawayThickness = temp;
			if (modelViewer)
			{
				modelViewer->setCutawayLineWidth((GLfloat)cutawayThickness);
			}
			TCUserDefaults::setLongForKey(cutawayThickness,
				CUTAWAY_THICKNESS_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hEffectsPage, IDC_STIPPLE,
			BM_GETCHECK, 0, 0) != 0) != useStipple)
		{
			useStipple = bTemp;
			if (modelViewer)
			{
				modelViewer->setUseStipple(useStipple);
			}
			TCUserDefaults::setLongForKey(useStipple, STIPPLE_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hEffectsPage, IDC_SORT,
			BM_GETCHECK, 0, 0) != 0) != sortTransparent)
		{
			sortTransparent = bTemp;
			if (modelViewer)
			{
				modelViewer->setSortTransparent(sortTransparent);
			}
			TCUserDefaults::setLongForKey(sortTransparent, SORT_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hEffectsPage, IDC_FLAT_SHADING,
			BM_GETCHECK, 0, 0) != 0) != usesFlatShading)
		{
			usesFlatShading = bTemp;
			if (modelViewer)
			{
				modelViewer->setUsesFlatShading(usesFlatShading);
			}
			TCUserDefaults::setLongForKey(usesFlatShading, FLAT_SHADING_KEY);
		}
		if ((bTemp = SendDlgItemMessage(hEffectsPage, IDC_SMOOTH_CURVES,
			BM_GETCHECK, 0, 0) != 0) != performSmoothing)
		{
			performSmoothing = bTemp;
			if (modelViewer)
			{
				modelViewer->setPerformSmoothing(performSmoothing);
			}
			TCUserDefaults::setLongForKey(performSmoothing,
				PERFORM_SMOOTHING_KEY);
		}
	}
}

void LDViewPreferences::applyPrimitivesChanges(void)
{
	if (hPrimitivesPage)
	{
		int temp;
		bool bTemp;

		if ((bTemp = SendDlgItemMessage(hPrimitivesPage,
			IDC_PRIMITIVE_SUBSTITUTION, BM_GETCHECK, 0, 0) != 0) !=
			allowPrimitiveSubstitution)
		{
			allowPrimitiveSubstitution = bTemp;
			if (modelViewer)
			{
				modelViewer->setAllowPrimitiveSubstitution(
					allowPrimitiveSubstitution);
			}
			TCUserDefaults::setLongForKey(allowPrimitiveSubstitution,
				PRIMITIVE_SUBSTITUTION_KEY);
		}
		if (allowPrimitiveSubstitution)
		{
			if ((bTemp = SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_STUDS,
				BM_GETCHECK, 0, 0) != 0) != textureStuds)
			{
				textureStuds = bTemp;
				if (modelViewer)
				{
					modelViewer->setTextureStuds(textureStuds);
				}
				TCUserDefaults::setLongForKey(textureStuds, TEXTURE_STUDS_KEY);
			}
			if (textureFilterType !=
				TCUserDefaults::longForKey(TEXTURE_FILTER_TYPE_KEY,
				GL_LINEAR_MIPMAP_LINEAR))
			{
				if (modelViewer)
				{
					modelViewer->setTextureFilterType(textureFilterType);
				}
				TCUserDefaults::setLongForKey(textureFilterType,
					TEXTURE_FILTER_TYPE_KEY);
			}
			else
			{
				textureFilterType =
					TCUserDefaults::longForKey(TEXTURE_FILTER_TYPE_KEY,
					GL_LINEAR_MIPMAP_LINEAR);
			}
			if ((temp = SendDlgItemMessage(hPrimitivesPage, IDC_CURVE_QUALITY,
				TBM_GETPOS, 0, 0)) != curveQuality)
			{
				curveQuality = temp;
				if (modelViewer)
				{
					modelViewer->setCurveQuality(curveQuality);
				}
				TCUserDefaults::setLongForKey(curveQuality, CURVE_QUALITY_KEY);
			}
		}
		if ((bTemp = SendDlgItemMessage(hPrimitivesPage, IDC_STUD_QUALITY,
			BM_GETCHECK, 0, 0) == 0) != qualityStuds)
		{
			qualityStuds = bTemp;
			if (modelViewer)
			{
				modelViewer->setQualityStuds(qualityStuds);
			}
			TCUserDefaults::setLongForKey(qualityStuds, QUALITY_STUDS_KEY);
		}
	}
}

void LDViewPreferences::applyChanges(void)
{
	applyGeneralChanges();
	applyGeometryChanges();
	applyEffectsChanges();
	applyPrimitivesChanges();
}

void LDViewPreferences::getRGB(int color, int &r, int &g, int &b)
{
	r = color & 0xFF;
	g = (color >> 8) & 0xFF;
	b = (color >> 16) & 0xFF;
}

void LDViewPreferences::chooseBackgroundColor(void)
{
	chooseColor(hBackgroundColorButton, hBackgroundColorBitmap,
		backgroundColor);
}

void LDViewPreferences::chooseDefaultColor(void)
{
	chooseColor(hDefaultColorButton, hDefaultColorBitmap, defaultColor);
}

void LDViewPreferences::chooseColor(HWND hColorButton, HBITMAP hColorBitmap,
									COLORREF &color)
{
	CHOOSECOLOR chooseColor;

	memset(&chooseColor, 0, sizeof CHOOSECOLOR);
	chooseColor.lStructSize = sizeof CHOOSECOLOR;
	chooseColor.hwndOwner = hPropSheet;
	chooseColor.rgbResult = color;
	chooseColor.lpCustColors = customColors;
	chooseColor.Flags = CC_ANYCOLOR | CC_RGBINIT;
	EnableWindow(hPropSheet, FALSE);
	if (ChooseColor(&chooseColor))
	{
		color = chooseColor.rgbResult;
		redrawColorBitmap(hColorButton, hColorBitmap, color);
	}
	EnableWindow(hPropSheet, TRUE);
}

void LDViewPreferences::doGeneralClick(int controlId, HWND /*controlHWnd*/)
{
	switch (controlId)
	{
		case IDC_FS_REFRESH:
			doFSRefresh();
			break;
		case IDC_BACKGROUND_COLOR:
			chooseBackgroundColor();
			break;
		case IDC_DEFAULT_COLOR:
			chooseDefaultColor();
			break;
		case IDC_FSAA:
			doFSAA();
			break;
		case IDC_FSAA_2X:
			fsaaMode = 2;
			break;
		case IDC_FSAA_2XP:
			fsaaMode = 3;
			break;
		case IDC_FSAA_4X:
			fsaaMode = 4;
			break;
		case IDC_FSAA_4XP:
			fsaaMode = 5;
			break;
		case IDC_GENERAL_RESET:
			loadDefaultGeneralSettings();
			setupGeneralPage();
			break;
	}
	enableApply(hGeneralPage);
}

void LDViewPreferences::doGeometryClick(int controlId, HWND /*controlHWnd*/)
{
	switch (controlId)
	{
		case IDC_HIGHLIGHTS:
			doHighlights();
			break;
		case IDC_WIREFRAME:
			doWireframe();
			break;
		case IDC_SEAMS:
			doSeams();
			break;
		case IDC_GEOMETRY_RESET:
			loadDefaultGeometrySettings();
			setupGeometryPage();
			break;
	}
	enableApply(hGeometryPage);
}

void LDViewPreferences::doEffectsClick(int controlId, HWND /*controlHWnd*/)
{
	switch (controlId)
	{
		case IDC_LIGHTING:
			doLighting();
			break;
		case IDC_STIPPLE:
			doStipple();
			break;
		case IDC_SORT:
			doSort();
			break;
		case IDC_FLAT_SHADING:
			doFlatShading();
			break;
		case IDC_SMOOTH_CURVES:
			doSmoothCurves();
			break;
		case IDC_STEREO:
			doStereo();
			break;
		case IDC_HARDWARE_STEREO:
			stereoMode = LDVStereoHardware;
			break;
		case IDC_CROSS_EYED_STEREO:
			stereoMode = LDVStereoCrossEyed;
			break;
		case IDC_PARALLEL_STEREO:
			stereoMode = LDVStereoParallel;
			break;
		case IDC_CUTAWAY:
			doCutaway();
			break;
		case IDC_CUTAWAY_COLOR:
			cutawayMode = LDVCutawayWireframe;
			setupOpacitySlider();
			break;
		case IDC_CUTAWAY_MONOCHROME:
			cutawayMode = LDVCutawayStencil;
			setupOpacitySlider();
			break;
		case IDC_EFFECTS_RESET:
			loadDefaultEffectsSettings();
			setupEffectsPage();
			break;
	}
	enableApply(hEffectsPage);
}

void LDViewPreferences::doPrimitivesClick(int controlId, HWND /*controlHWnd*/)
{
	switch (controlId)
	{
		case IDC_PRIMITIVE_SUBSTITUTION:
			doPrimitives();
			break;
		case IDC_TEXTURE_STUDS:
			doTextureStuds();
			break;
		case IDC_TEXTURE_NEAREST:
			textureFilterType = GL_NEAREST_MIPMAP_NEAREST;
			break;
		case IDC_TEXTURE_BILINEAR:
			textureFilterType = GL_LINEAR_MIPMAP_NEAREST;
			break;
		case IDC_TEXTURE_TRILINEAR:
			textureFilterType = GL_LINEAR_MIPMAP_LINEAR;
			break;
		case IDC_PRIMITIVES_RESET:
			loadDefaultPrimitivesSettings();
			setupPrimitivesPage();
			break;
	}
	enableApply(hPrimitivesPage);
}

DWORD LDViewPreferences::doClick(HWND hPage, int controlId, HWND controlHWnd)
{
	bool wasPaused = true;

	if (modelViewer)
	{
		wasPaused = modelViewer->getPaused() != 0;
		if (!wasPaused)
		{
			modelViewer->pause();
		}
	}
	if (hPage == hGeneralPage)
	{
		doGeneralClick(controlId, controlHWnd);
	}
	else if (hPage == hGeometryPage)
	{
		doGeometryClick(controlId, controlHWnd);
	}
	else if (hPage == hEffectsPage)
	{
		doEffectsClick(controlId, controlHWnd);
	}
	else if (hPage == hPrimitivesPage)
	{
		doPrimitivesClick(controlId, controlHWnd);
	}
	if (!wasPaused)
	{
		modelViewer->unpause();
	}
	return 0;
}

void LDViewPreferences::doFSRefresh(void)
{
	if (SendDlgItemMessage(hGeneralPage, IDC_FS_REFRESH, BM_GETCHECK, 0, 0))
	{
		EnableWindow(hFullScreenRateField, TRUE);
	}
	else
	{
		SendDlgItemMessage(hGeneralPage, IDC_FS_RATE, WM_SETTEXT, 0,
			(LPARAM)"");
		EnableWindow(hFullScreenRateField, FALSE);
	}
}

void LDViewPreferences::doStipple(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_STIPPLE, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hEffectsPage, IDC_SORT, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doSort(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_SORT, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hEffectsPage, IDC_STIPPLE, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doFlatShading(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_FLAT_SHADING, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hEffectsPage, IDC_SMOOTH_CURVES, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doSmoothCurves(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_SMOOTH_CURVES, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hEffectsPage, IDC_FLAT_SHADING, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doHighlights(void)
{
	if (SendDlgItemMessage(hGeometryPage, IDC_HIGHLIGHTS, BM_GETCHECK, 0, 0))
	{
		enableEdges();
	}
	else
	{
		disableEdges();
	}
}

void LDViewPreferences::doWireframe(void)
{
	if (SendDlgItemMessage(hGeometryPage, IDC_WIREFRAME, BM_GETCHECK, 0, 0))
	{
		enableWireframe();
	}
	else
	{
		disableWireframe();
	}
}

void LDViewPreferences::doLighting(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_LIGHTING, BM_GETCHECK, 0, 0))
	{
		enableLighting();
	}
	else
	{
		disableLighting();
	}
}

void LDViewPreferences::doStereo(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_STEREO, BM_GETCHECK, 0, 0))
	{
		stereoMode = LDVStereoCrossEyed;
		enableStereo();
	}
	else
	{
		stereoMode = LDVStereoNone;
		disableStereo();
	}
}

void LDViewPreferences::doCutaway(void)
{
	if (SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY, BM_GETCHECK, 0,
		0))
	{
		cutawayMode = LDVCutawayWireframe;
		enableCutaway();
	}
	else
	{
		cutawayMode = LDVCutawayNormal;
		disableCutaway();
	}
}

void LDViewPreferences::doFSAA(void)
{
	if (SendDlgItemMessage(hGeneralPage, IDC_FSAA, BM_GETCHECK, 0, 0))
	{
		fsaaMode = 4;
		enableFSAA();
	}
	else
	{
		fsaaMode = 0;
		disableFSAA();
	}
}

void LDViewPreferences::doSeams(void)
{
	if (SendDlgItemMessage(hGeometryPage, IDC_SEAMS, BM_GETCHECK, 0, 0))
	{
		enableSeams();
	}
	else
	{
		disableSeams();
	}
}

void LDViewPreferences::doPrimitives(void)
{
	if (SendDlgItemMessage(hPrimitivesPage, IDC_PRIMITIVE_SUBSTITUTION,
		BM_GETCHECK, 0, 0))
	{
		enablePrimitives();
	}
	else
	{
		disablePrimitives();
	}
}

void LDViewPreferences::doTextureStuds(void)
{
	if (SendDlgItemMessage(hPrimitivesPage, IDC_TEXTURE_STUDS, BM_GETCHECK, 0,
		0))
	{
		enableTextureFiltering();
	}
	else
	{
		disableTextureFiltering();
	}
}

void LDViewPreferences::doReset(void)
{
	loadSettings();
/*
	backgroundColor =
		(COLORREF)TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY);
	defaultColor = (COLORREF)TCUserDefaults::longForKey(DEFAULT_COLOR_KEY,
		0x999999);
	seamWidth = TCUserDefaults::longForKey(SEAM_WIDTH_KEY);
	fsaaMode = TCUserDefaults::longForKey(FSAA_MODE_KEY);
	stereoMode = (LDVStereoMode)TCUserDefaults::longForKey(STEREO_MODE_KEY);
	cutawayMode = (LDVCutawayMode)TCUserDefaults::longForKey(CUTAWAY_MODE_KEY);
	textureFilterType = TCUserDefaults::longForKey(TEXTURE_FILTER_TYPE_KEY,
		GL_LINEAR_MIPMAP_LINEAR);
*/
}

BOOL LDViewPreferences::doDialogVScroll(HWND hDlg, int scrollCode, int position,
										HWND hScrollBar)
{
	if (scrollCode == SB_THUMBPOSITION && hScrollBar == hSeamSpin)
	{
		char seamWidthString[128];

		seamWidth += (position - seamWidth);
		sprintf(seamWidthString, "%0.2f", seamWidth / 100.0f);
		SendDlgItemMessage(hDlg, IDC_SEAM_WIDTH_FIELD, WM_SETTEXT, 0,
			(LPARAM)seamWidthString);
		SendDlgItemMessage(hDlg, IDC_SEAM_SPIN, UDM_SETPOS, 0, seamWidth);
		if (seamWidth)
		{
			SendDlgItemMessage(hGeometryPage, IDC_SEAMS, BM_SETCHECK,
				1, 0);
		}
		else
		{
			SendDlgItemMessage(hGeometryPage, IDC_SEAMS, BM_SETCHECK,
				0, 0);
		}
		enableApply(hGeometryPage);
	}
	return TRUE;
}

void LDViewPreferences::setupPage(int pageNumber)
{
	if (pageNumber == generalPageNumber)
	{
		setupGeneralPage();
	}
	else if (pageNumber == geometryPageNumber)
	{
		setupGeometryPage();
	}
	else if (pageNumber == effectsPageNumber)
	{
		setupEffectsPage();
	}
	else if (pageNumber == primitivesPageNumber)
	{
		setupPrimitivesPage();
	}
	disableApply(hwndArray->pointerAtIndex(pageNumber));
}

void LDViewPreferences::setupGeneralPage(void)
{
	hGeneralPage = hwndArray->pointerAtIndex(generalPageNumber);
	setupFSAA();
	SendDlgItemMessage(hGeneralPage, IDC_LINE_AA, BM_SETCHECK, lineSmoothing,
		0);
	SendDlgItemMessage(hGeneralPage, IDC_FRAME_RATE, BM_SETCHECK, showsFPS,
		0);
	SendDlgItemMessage(hGeneralPage, IDC_SHOW_ERRORS, BM_SETCHECK,
		showErrors, 0);
	setupFullScreenRefresh();
	setupBackgroundColorButton();
	setupDefaultColorButton();
}

void LDViewPreferences::setupModelGeometry(void)
{
//	hCurveQualityLabel = GetDlgItem(hGeometryPage, IDC_CURVE_QUALITY_LABEL);
//	hCurveQualitySlider = GetDlgItem(hGeometryPage, IDC_CURVE_QUALITY);
//	SendDlgItemMessage(hGeometryPage, IDC_STUD_QUALITY, BM_SETCHECK,
//		!qualityStuds, 0);
//	setupDialogSlider(hGeometryPage, IDC_CURVE_QUALITY, 1, 6, 1,
//		curveQuality);
	setupSeamWidth();
/*
	if (allowPrimitiveSubstitution)
	{
		enablePrimitives();
	}
	else
	{
		disablePrimitives();
	}
*/
}

void LDViewPreferences::enableWireframe(void)
{
	EnableWindow(hWireframeFogButton, TRUE);
	EnableWindow(hWireframeThicknessLabel, TRUE);
	EnableWindow(hWireframeThicknessSlider, TRUE);
}

void LDViewPreferences::disableWireframe(void)
{
	EnableWindow(hWireframeFogButton, FALSE);
	EnableWindow(hWireframeThicknessLabel, FALSE);
	EnableWindow(hWireframeThicknessSlider, FALSE);
}

void LDViewPreferences::setupWireframe(void)
{
	hWireframeFogButton = GetDlgItem(hGeometryPage, IDC_WIREFRAME_FOG);
	hWireframeThicknessLabel = GetDlgItem(hGeometryPage,
		IDC_WIREFRAME_THICKNESS_LABEL);
	hWireframeThicknessSlider = GetDlgItem(hGeometryPage,
		IDC_WIREFRAME_THICKNESS);
	SendDlgItemMessage(hGeometryPage, IDC_WIREFRAME, BM_SETCHECK, drawWireframe,
		0);
	SendDlgItemMessage(hGeometryPage, IDC_WIREFRAME_FOG, BM_SETCHECK,
		useWireframeFog, 0);
	setupDialogSlider(hGeometryPage, IDC_WIREFRAME_THICKNESS, 1, 5, 1,
		wireframeThickness);
	if (drawWireframe)
	{
		enableWireframe();
	}
	else
	{
		disableWireframe();
	}
}

void LDViewPreferences::enableEdges(void)
{
	EnableWindow(hConditionalHighlightsButton, TRUE);
	EnableWindow(hQualityLinesButton, TRUE);
	EnableWindow(hAlwaysBlackButton, TRUE);
	EnableWindow(hEdgeThicknessLabel, TRUE);
	EnableWindow(hEdgeThicknessSlider, TRUE);
	SendDlgItemMessage(hGeometryPage, IDC_CONDITIONAL_HIGHLIGHTS, BM_SETCHECK,
		drawConditionalHighlights ? 1 : 0, 0);
	SendDlgItemMessage(hGeometryPage, IDC_QUALITY_LINES, BM_SETCHECK,
		usePolygonOffset, 0);
	SendDlgItemMessage(hGeometryPage, IDC_ALWAYS_BLACK, BM_SETCHECK,
		blackHighlights, 0);
}

void LDViewPreferences::disableEdges(void)
{
	EnableWindow(hConditionalHighlightsButton, FALSE);
	EnableWindow(hQualityLinesButton, FALSE);
	EnableWindow(hAlwaysBlackButton, FALSE);
	EnableWindow(hEdgeThicknessLabel, FALSE);
	EnableWindow(hEdgeThicknessSlider, FALSE);
	SendDlgItemMessage(hGeometryPage, IDC_CONDITIONAL_HIGHLIGHTS, BM_SETCHECK,
		0, 0);
	SendDlgItemMessage(hGeometryPage, IDC_QUALITY_LINES, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeometryPage, IDC_ALWAYS_BLACK, BM_SETCHECK, 0, 0);
}

void LDViewPreferences::setupEdgeLines(void)
{
	hConditionalHighlightsButton = GetDlgItem(hGeometryPage,
		IDC_CONDITIONAL_HIGHLIGHTS);
	hQualityLinesButton = GetDlgItem(hGeometryPage, IDC_QUALITY_LINES);
	hAlwaysBlackButton = GetDlgItem(hGeometryPage, IDC_ALWAYS_BLACK);
	hEdgeThicknessLabel = GetDlgItem(hGeometryPage, IDC_EDGE_THICKNESS_LABEL);
	hEdgeThicknessSlider = GetDlgItem(hGeometryPage, IDC_EDGE_THICKNESS);
	SendDlgItemMessage(hGeometryPage, IDC_HIGHLIGHTS, BM_SETCHECK,
		showsHighlightLines, 0);
	setupDialogSlider(hGeometryPage, IDC_EDGE_THICKNESS, 1, 5, 1,
		edgeThickness);
	if (showsHighlightLines)
	{
		enableEdges();
	}
	else
	{
		disableEdges();
	}
}

void LDViewPreferences::setupGeometryPage(void)
{
	hGeometryPage = hwndArray->pointerAtIndex(geometryPageNumber);
	setupModelGeometry();
	setupWireframe();
	setupEdgeLines();
}

void LDViewPreferences::setupOpacitySlider(void)
{
	if (cutawayMode == LDVCutawayWireframe && !LDVExtensionsSetup::haveAlpha())
	{
		cutawayAlpha = 100;
		EnableWindow(hCutawayOpacitySlider, FALSE);
		EnableWindow(hCutawayOpacityLabel, FALSE);
		SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_OPACITY, TBM_SETPOS, 1,
			100);
	}
	else
	{
		EnableWindow(hCutawayOpacitySlider, TRUE);
		EnableWindow(hCutawayOpacityLabel, TRUE);
	}
}

void LDViewPreferences::enableCutaway(void)
{
	int activeCutaway = 0;

	SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY, BM_SETCHECK, 1, 0);
	EnableWindow(hCutawayColorButton, TRUE);
	if (LDVExtensionsSetup::haveStencil())
	{
		EnableWindow(hCutawayMonochromButton, TRUE);
	}
	else
	{
		if (cutawayMode == LDVCutawayStencil)
		{
			cutawayMode = LDVCutawayWireframe;
		}
		EnableWindow(hCutawayMonochromButton, FALSE);
	}
	setupOpacitySlider();
	EnableWindow(hCutawayThicknessSlider, TRUE);
	EnableWindow(hCutawayThicknessLabel, TRUE);
	SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_COLOR, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_MONOCHROME, BM_SETCHECK, 0, 0);
	switch (cutawayMode)
	{
	case LDVCutawayWireframe:
		activeCutaway = IDC_CUTAWAY_COLOR;
		break;
	case LDVCutawayStencil:
		activeCutaway = IDC_CUTAWAY_MONOCHROME;
		break;
	}
	SendDlgItemMessage(hEffectsPage, activeCutaway, BM_SETCHECK, 1, 0);
}

void LDViewPreferences::disableCutaway(void)
{
	SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY, BM_SETCHECK, 0, 0);
	EnableWindow(hCutawayColorButton, FALSE);
	EnableWindow(hCutawayMonochromButton, FALSE);
	EnableWindow(hCutawayOpacityLabel, FALSE);
	EnableWindow(hCutawayOpacitySlider, FALSE);
	EnableWindow(hCutawayThicknessLabel, FALSE);
	EnableWindow(hCutawayThicknessSlider, FALSE);
	SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_COLOR, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_CUTAWAY_MONOCHROME, BM_SETCHECK, 0, 0);
}

void LDViewPreferences::setupCutaway(void)
{
	hCutawayColorButton = GetDlgItem(hEffectsPage, IDC_CUTAWAY_COLOR);
	hCutawayMonochromButton = GetDlgItem(hEffectsPage, IDC_CUTAWAY_MONOCHROME);
	hCutawayOpacityLabel = GetDlgItem(hEffectsPage, IDC_CUTAWAY_OPACITY_LABEL);
	hCutawayOpacitySlider = GetDlgItem(hEffectsPage, IDC_CUTAWAY_OPACITY);
	hCutawayThicknessLabel = GetDlgItem(hEffectsPage,
		IDC_CUTAWAY_THICKNESS_LABEL);
	hCutawayThicknessSlider = GetDlgItem(hEffectsPage, IDC_CUTAWAY_THICKNESS);
	setupDialogSlider(hEffectsPage, IDC_CUTAWAY_OPACITY, 1, 100, 10,
		cutawayAlpha);
	setupDialogSlider(hEffectsPage, IDC_CUTAWAY_THICKNESS, 1, 5, 1,
		cutawayThickness);
	if (cutawayMode == LDVCutawayNormal)
	{
		disableCutaway();
	}
	else
	{
		enableCutaway();
	}
}

void LDViewPreferences::enableStereo(void)
{
	int activeStereo = 0;

	SendDlgItemMessage(hEffectsPage, IDC_STEREO, BM_SETCHECK, 1, 0);
	EnableWindow(hHardwareStereoButton, TRUE);
	EnableWindow(hCrossEyedStereoButton, TRUE);
	EnableWindow(hParallelStereoButton, TRUE);
	EnableWindow(hStereoSpacingSlider, TRUE);
	EnableWindow(hStereoSpacingLabel, TRUE);
	SendDlgItemMessage(hEffectsPage, IDC_HARDWARE_STEREO, BM_SETCHECK, 0,
		0);
	SendDlgItemMessage(hEffectsPage, IDC_CROSS_EYED_STEREO, BM_SETCHECK, 0,
		0);
	SendDlgItemMessage(hEffectsPage, IDC_PARALLEL_STEREO, BM_SETCHECK, 0,
		0);
	switch (stereoMode)
	{
	case LDVStereoHardware:
		activeStereo = IDC_HARDWARE_STEREO;
		break;
	case LDVStereoCrossEyed:
		activeStereo = IDC_CROSS_EYED_STEREO;
		break;
	case LDVStereoParallel:
		activeStereo = IDC_PARALLEL_STEREO;
		break;
	}
	SendDlgItemMessage(hEffectsPage, activeStereo, BM_SETCHECK, 1, 0);
}

void LDViewPreferences::disableStereo(void)
{
	SendDlgItemMessage(hEffectsPage, IDC_STEREO, BM_SETCHECK, 0, 0);
	EnableWindow(hHardwareStereoButton, FALSE);
	EnableWindow(hCrossEyedStereoButton, FALSE);
	EnableWindow(hParallelStereoButton, FALSE);
	EnableWindow(hStereoSpacingSlider, FALSE);
	EnableWindow(hStereoSpacingLabel, FALSE);
	SendDlgItemMessage(hEffectsPage, IDC_HARDWARE_STEREO, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_CROSS_EYED_STEREO, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_PARALLEL_STEREO, BM_SETCHECK, 0, 0);
}

void LDViewPreferences::setupStereo(void)
{
	hHardwareStereoButton = GetDlgItem(hEffectsPage, IDC_HARDWARE_STEREO);
	hCrossEyedStereoButton = GetDlgItem(hEffectsPage, IDC_CROSS_EYED_STEREO);
	hParallelStereoButton = GetDlgItem(hEffectsPage, IDC_PARALLEL_STEREO);
	hStereoSpacingSlider = GetDlgItem(hEffectsPage, IDC_STEREO_SPACING);
	hStereoSpacingLabel = GetDlgItem(hEffectsPage, IDC_STEREO_SPACING_LABEL);
	setupDialogSlider(hEffectsPage, IDC_STEREO_SPACING, 0, 100, 10,
		stereoEyeSpacing);
	if (stereoMode == LDVStereoNone)
	{
		disableStereo();
	}
	else
	{
		enableStereo();
	}
}

void LDViewPreferences::enableLighting(void)
{
	EnableWindow(hLightQualityButton, TRUE);
	EnableWindow(hLightSubduedButton, TRUE);
	EnableWindow(hLightSpecularButton, TRUE);
	EnableWindow(hLightAlternateButton, TRUE);
	SendDlgItemMessage(hEffectsPage, IDC_LIGHTING, BM_SETCHECK, 1, 0);
	SendDlgItemMessage(hEffectsPage, IDC_LIGHTING_QUALITY, BM_SETCHECK,
		qualityLighting, 0);
	SendDlgItemMessage(hEffectsPage, IDC_LIGHTING_SUBDUED, BM_SETCHECK,
		subduedLighting, 0);
	SendDlgItemMessage(hEffectsPage, IDC_SPECULAR, BM_SETCHECK,
		usesSpecular, 0);
	SendDlgItemMessage(hEffectsPage, IDC_ALTERNATE_LIGHTING, BM_SETCHECK,
		oneLight, 0);
}

void LDViewPreferences::disableLighting(void)
{
	EnableWindow(hLightQualityButton, FALSE);
	EnableWindow(hLightSubduedButton, FALSE);
	EnableWindow(hLightSpecularButton, FALSE);
	EnableWindow(hLightAlternateButton, FALSE);
	SendDlgItemMessage(hEffectsPage, IDC_LIGHTING, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_LIGHTING_QUALITY, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_LIGHTING_SUBDUED, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_SPECULAR, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hEffectsPage, IDC_ALTERNATE_LIGHTING, BM_SETCHECK, 0, 0);
}

void LDViewPreferences::setupLighting(void)
{
	hLightQualityButton = GetDlgItem(hEffectsPage, IDC_LIGHTING_QUALITY);
	hLightSubduedButton = GetDlgItem(hEffectsPage, IDC_LIGHTING_SUBDUED);
	hLightSpecularButton = GetDlgItem(hEffectsPage, IDC_SPECULAR);
	hLightAlternateButton = GetDlgItem(hEffectsPage, IDC_ALTERNATE_LIGHTING);
	if (useLighting)
	{
		enableLighting();
	}
	else
	{
		disableLighting();
	}
}

void LDViewPreferences::setupEffectsPage(void)
{
	hEffectsPage = hwndArray->pointerAtIndex(effectsPageNumber);
	setupLighting();
	setupStereo();
	setupCutaway();
	SendDlgItemMessage(hEffectsPage, IDC_STIPPLE, BM_SETCHECK,
		useStipple, 0);
	SendDlgItemMessage(hEffectsPage, IDC_SORT, BM_SETCHECK,
		sortTransparent, 0);
	SendDlgItemMessage(hEffectsPage, IDC_FLAT_SHADING, BM_SETCHECK,
		usesFlatShading, 0);
	SendDlgItemMessage(hEffectsPage, IDC_SMOOTH_CURVES, BM_SETCHECK,
		performSmoothing, 0);
}

void LDViewPreferences::setupSubstitution(void)
{
	hTextureStudsButton = GetDlgItem(hPrimitivesPage, IDC_TEXTURE_STUDS);
	hTextureNearestButton = GetDlgItem(hPrimitivesPage, IDC_TEXTURE_NEAREST);
	hTextureBilinearButton = GetDlgItem(hPrimitivesPage, IDC_TEXTURE_BILINEAR);
	hTextureTrilinearButton = GetDlgItem(hPrimitivesPage,
		IDC_TEXTURE_TRILINEAR);
	hCurveQualityLabel = GetDlgItem(hPrimitivesPage, IDC_CURVE_QUALITY_LABEL);
	hCurveQualitySlider = GetDlgItem(hPrimitivesPage, IDC_CURVE_QUALITY);
	setupDialogSlider(hPrimitivesPage, IDC_CURVE_QUALITY, 1, 12, 1,
		curveQuality);
	if (allowPrimitiveSubstitution)
	{
		enablePrimitives();
	}
	else
	{
		disablePrimitives();
	}
}

void LDViewPreferences::setupPrimitivesPage(void)
{
	hPrimitivesPage = hwndArray->pointerAtIndex(primitivesPageNumber);
	setupSubstitution();
	SendDlgItemMessage(hPrimitivesPage, IDC_STUD_QUALITY, BM_SETCHECK,
		!qualityStuds, 0);
}

void LDViewPreferences::enableFSAA(void)
{
	int activeFSAA;

	SendDlgItemMessage(hGeneralPage, IDC_FSAA, BM_SETCHECK, 1, 0);
	EnableWindow(hFSAA2xButton, TRUE);
	EnableWindow(hFSAA4xButton, TRUE);
	if (LDVExtensionsSetup::haveQuincunxExtension())
	{
		EnableWindow(hFSAA2xEnhancedButton, TRUE);
		EnableWindow(hFSAA4xEnhancedButton, TRUE);
	}
	else
	{
		EnableWindow(hFSAA2xEnhancedButton, FALSE);
		EnableWindow(hFSAA4xEnhancedButton, FALSE);
		if (fsaaMode & 1)
		{
			fsaaMode--;
			TCUserDefaults::setLongForKey(fsaaMode, FSAA_MODE_KEY);
		}
	}
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_2X, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_4X, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_2XP, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_4XP, BM_SETCHECK, 0, 0);
	switch (fsaaMode)
	{
	case 2:
		activeFSAA = IDC_FSAA_2X;
		break;
	case 3:
		activeFSAA = IDC_FSAA_2XP;
		break;
	case 4:
		activeFSAA = IDC_FSAA_4X;
		break;
	case 5:
		activeFSAA = IDC_FSAA_4XP;
		break;
	default:
		fsaaMode = 4;
		activeFSAA = IDC_FSAA_4X;
		break;
	}
	SendDlgItemMessage(hGeneralPage, activeFSAA, BM_SETCHECK, 1, 0);
}

void LDViewPreferences::disableFSAA(void)
{
	SendDlgItemMessage(hGeneralPage, IDC_FSAA, BM_SETCHECK, 0, 0);
	EnableWindow(hFSAA2xButton, FALSE);
	EnableWindow(hFSAA2xEnhancedButton, FALSE);
	EnableWindow(hFSAA4xButton, FALSE);
	EnableWindow(hFSAA4xEnhancedButton, FALSE);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_2X, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_2XP, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_4X, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_4XP, BM_SETCHECK, 0, 0);
}

void LDViewPreferences::setupFSAA(void)
{
	bool haveMultisampleExtension =
		LDVExtensionsSetup::haveMultisampleExtension();

	if (!haveMultisampleExtension)
	{
		fsaaMode = 0;
		EnableWindow(GetDlgItem(hGeneralPage, IDC_FSAA), FALSE);
	}
	hFSAA2xButton = GetDlgItem(hGeneralPage, IDC_FSAA_2X);
	hFSAA2xEnhancedButton = GetDlgItem(hGeneralPage, IDC_FSAA_2XP);
	hFSAA4xButton = GetDlgItem(hGeneralPage, IDC_FSAA_4X);
	hFSAA4xEnhancedButton = GetDlgItem(hGeneralPage, IDC_FSAA_4XP);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_2X, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_2XP, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_4X, BM_SETCHECK, 0, 0);
	SendDlgItemMessage(hGeneralPage, IDC_FSAA_4XP, BM_SETCHECK, 0, 0);
	if (fsaaMode)
	{
		enableFSAA();
	}
	else
	{
		disableFSAA();
	}
}

bool LDViewPreferences::doApply(void)
{
	applyChanges();
	return true;
}

void LDViewPreferences::clear(void)
{
	if (hBackgroundColorBitmap)
	{
		DeleteObject(hBackgroundColorBitmap);
		hBackgroundColorBitmap = NULL;
	}
	if (hDefaultColorBitmap)
	{
		DeleteObject(hDefaultColorBitmap);
		hDefaultColorBitmap = NULL;
	}
	if (hButtonColorDC)
	{
		DeleteDC(hButtonColorDC);
		hButtonColorDC = NULL;
	}
	hGeneralPage = NULL;
	hGeometryPage = NULL;
	hEffectsPage = NULL;
	hPrimitivesPage = NULL;
	CUIPropertySheet::clear();
}
